@@ -462,6 +462,7 @@ find_nxt_kdc(struct tr_state *ts)
 	if (ts->ntgts > 0) {
 	    /* Punt NXT_TGT from KDC_TGTS if bogus. */
 	    krb5_free_creds(ts->ctx, ts->kdc_tgts[--ts->ntgts]);
+	    ts->kdc_tgts[ts->ntgts] = NULL;
 	}
 	TR_DBG_RET(ts, "find_nxt_kdc", KRB5_KDCREP_MODIFIED);
 	return KRB5_KDCREP_MODIFIED;
@@ -1042,6 +1043,7 @@ krb5_get_cred_from_kdc_opt(krb5_context context, krb5_ccache ccache,
 	    krb5_free_creds(context, (*tgts)[i]);
 	}
 	free(*tgts);
+	*tgts = NULL;
     }
     context->use_conf_ktypes = 1;
     retval = krb5_cc_retrieve_cred(context, ccache, RETR_FLAGS,
