@@ -52,7 +52,7 @@
  */
 #define KRB5_MAJOR_RELEASE 1
 #define KRB5_MINOR_RELEASE 6
-#define KRB5_PATCHLEVEL 0
+#define KRB5_PATCHLEVEL 2
 /* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20070109"
-#define KRB5_RELTAG "tags/krb5-1-6-final"
+#define KRB5_RELDATE "20070705"
+#define KRB5_RELTAG "tags/krb5-1-6-2-final"
